from setuptools import find_packages, setup

setup(
    name='danling',
    version='0.0.1a3',
    description='danling',
    author='Zhiyuan Chen',
    author_email='this@zyc.ai',
    url='http://github.com/ZhiyuanChen/danling',
    packages=find_packages(),
)
